﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Shapes;
using Infragistics.Controls.Charts;

namespace Infragistics.Extension
{
    /// <summary>
    /// Represents a custom type of CustomScatterSeries for XamDataChart control.
    /// </summary>
    public class CustomScatterSeries : Series //MarkerSeries //Series
    {
        public CustomScatterSeries()
        {
            this.DefaultStyleKey = typeof(CustomScatterSeries);
        }
        
        protected KDTree<object> ItemsTree;
        
        #region Layers Support

        /// <summary>
        /// If possible, will return the best available position within the series that has the best value match for the world position provided. 
        /// </summary>
        /// <param name="world">The world coordinate for which to get a value position for</param>
        /// <param name="useInterpolation">If true, interpolation should be used to get in-between values, rather than only the actual values in the data set.</param>
        /// <param name="skipUnknowns">If true, unknown values will be skipped when searching for a series value.</param>
        /// <returns>The position in question, or NaN,NaN if no valid position is available.</returns>
        public override Point GetSeriesValuePosition(Point world, bool useInterpolation, bool skipUnknowns)
        {
         

            var seriesPoint = new Point(double.NaN, double.NaN);
            if (this.ItemsTree == null)
                return seriesPoint;

            if (!world.IsPlottable()) return seriesPoint;

            var position = GetSeriesExactPosition(world);
            var neighbors = this.ItemsTree.GetNeighbors(3, position.X, position.Y).ToList();
           
            if (neighbors.Any())
            {
                Point closestPoint;
  
                var itemA = neighbors[0];
                var pointA = new Point(itemA.Point[0], itemA.Point[1]);

                if (useInterpolation && neighbors.Count() >= 2)
                {
                    var itemB = neighbors[1];
                    var pointB = new Point(itemB.Point[0], itemB.Point[1]);
                    closestPoint = position.GetIntersectionWith(pointA, pointB);

                    if (!closestPoint.IsOnSegment(pointA, pointB))
                    {
                        closestPoint = pointA;
                    }
                }
                else
                {
                    closestPoint = pointA; 
                }

                seriesPoint = GetSeriesScaledPosition(closestPoint);
   
            }
           
            return seriesPoint;

        }

        /// <summary>
        /// Gets the item item associated with the specified world position 
        /// where a tooltip must be displayed on the series' marker
        /// </summary>
        public override object GetItem(Point world)
        {
            //var seriesPoint = new Point(double.NaN, double.NaN);
            if (this.ItemsTree == null)
                return null;

            var position = GetSeriesExactPosition(world);
            var neighbors = this.ItemsTree.GetNeighbors(2, position.X, position.Y);

            var closest = neighbors.FirstOrDefault();
            if (closest != null)
            {
                // var x = closest.Point[0];
                //var y = closest.Point[1];
                var item = closest.Value;
                return item;
                //seriesPoint = GetSeriesScaledPosition(new Point(x, y));
            }


            return null;
        }
       
        #endregion


        public override bool IsScatter { get  { return true; } }
        public override bool IsCategory { get { return false; } }
        

        #region Properties

        internal const string ItemsMemberPathXPropertyName = "ItemsMemberPathX";
        public static readonly DependencyProperty ItemsMemberPathXProperty = DependencyProperty.Register(
            ItemsMemberPathXPropertyName, typeof(string), typeof(CustomScatterSeries),
            new PropertyMetadata("X", (sender, e) =>
            {
                var series = (CustomScatterSeries)sender;
                series.RaisePropertyChanged(ItemsMemberPathXPropertyName, e.OldValue, e.NewValue);
            }));


        /// <summary>
        /// Gets or sets x-value member path for data items
        /// </summary>
        public string ItemsMemberPathX
        {
            get { return (string)GetValue(ItemsMemberPathXProperty); }
            set { SetValue(ItemsMemberPathXProperty, value); }
        }

        internal const string ItemsMemberPathYPropertyName = "ItemsMemberPathY";
        public static readonly DependencyProperty ItemsMemberPathYProperty = DependencyProperty.Register(
            ItemsMemberPathYPropertyName, typeof(string), typeof(CustomScatterSeries),
            new PropertyMetadata("Y", (sender, e) =>
            {
                var series = (CustomScatterSeries)sender;
                series.RaisePropertyChanged(ItemsMemberPathYPropertyName, e.OldValue, e.NewValue);
            }));


        /// <summary>
        /// Gets or sets y-value member path for data items
        /// </summary>
        public string ItemsMemberPathY
        {
            get { return (string)GetValue(ItemsMemberPathYProperty); }
            set { SetValue(ItemsMemberPathYProperty, value); }
        }


        #region Property - XAxis

        public const string XAxisPropertyName = "XAxis";
        public static readonly DependencyProperty XAxisProperty =
            DependencyProperty.Register(XAxisPropertyName, typeof(NumericXAxis),
            typeof(CustomScatterSeries), new PropertyMetadata(null, (sender, e) =>
            {
                var series = (CustomScatterSeries)sender;
                series.RaisePropertyChanged(XAxisPropertyName, e.OldValue, e.NewValue);
            }));


        /// <summary>
        /// Gets or sets x-axis 
        /// </summary>
        public NumericXAxis XAxis
        {
            get { return this.GetValue(XAxisProperty) as NumericXAxis; }
            set { this.SetValue(XAxisProperty, value); }
        }

        #endregion



        #region Property - YAxis

        public const string YAxisPropertyName = "YAxis";
        public static readonly DependencyProperty YAxisProperty =
            DependencyProperty.Register(YAxisPropertyName, typeof(NumericYAxis),
            typeof(CustomScatterSeries), new PropertyMetadata(null, (sender, e) =>
            {
                var series = (CustomScatterSeries)sender;
                series.RaisePropertyChanged(YAxisPropertyName, e.OldValue, e.NewValue);
            }));


        /// <summary>
        /// Gets or sets y-axis 
        /// </summary>
        public NumericYAxis YAxis
        {
            get { return this.GetValue(YAxisProperty) as NumericYAxis; }
            set { this.SetValue(YAxisProperty, value); }
        }


        #endregion
        

        #endregion
   
      

        #region Overriden Methods


        /// <summary>
        /// Renders the Custom Series
        /// </summary>
        protected override void RenderSeriesOverride(bool animate)
        {
            base.RenderSeriesOverride(animate);
            // check if the series can be rendered:
            // - the Viewport (the bounds rectangle for the series) is not empty, 
            // - the RootCanvas (the container for the custom graphics) is not null.  
            // - the Axes are not null.  
            // - the ItemsSource is not null.  
            if (this.Viewport.IsEmpty || this.RootCanvas == null ||
                this.XAxis == null || this.YAxis == null ||
                this.ItemsSource == null)
            {
                return;
            }

            // clears the RootCanvas on every render of the series
            this.RootCanvas.Children.Clear();

            // create data structure for the Custom Series based on values of items in the source of this series
            var itemsDictionary = new Dictionary<double, PointCollection>();
            if (this.ItemsTree == null || this.ItemsTree.Count == 0)
                return;

            var first = ItemsSource.Cast<object>().First();

            PathFigure figure = new PathFigure();

            var dataPoint2 = GetDataPoint(first);

            var param2 = new ScalerParams(this.SeriesViewer.WindowRect, this.Viewport, false);

            var xPoint = this.XAxis.GetScaledValue(dataPoint2.X, param2);
            var yPoint = this.YAxis.GetScaledValue(dataPoint2.Y, param2);

            figure.StartPoint = new Point(xPoint, yPoint);

            PolyLineSegment segment = new PolyLineSegment();       

            foreach (var dataItem in this.ItemsSource)
            {

                var dataPoint = GetDataPoint(dataItem);

                var isInverted = this.XAxis.IsInverted;
                var param = new ScalerParams(this.SeriesViewer.WindowRect, this.Viewport, isInverted);
                // scale x-locations of data point to the series' viewport
                var x = this.XAxis.GetScaledValue(dataPoint.X, param);

                isInverted = this.YAxis.IsInverted;
                param = new ScalerParams(this.SeriesViewer.WindowRect, this.Viewport, isInverted);
                // scale y-location of data point to the series' viewport
                var y = this.YAxis.GetScaledValue(dataPoint.Y, param);

                segment.Points.Add(new Point(x, y));

            }

            PathSegmentCollection col123 = new PathSegmentCollection();
            col123.Add(segment);
            figure.Segments = col123;

            PathFigureCollection figCol = new PathFigureCollection();
            figCol.Add(figure);

            PathGeometry geometry = new PathGeometry();
            geometry.Figures = figCol;

            Path p = new Path();

            p.Stroke = Brushes.Red;
            p.StrokeThickness = 1;
            p.Data = geometry;

            this.RootCanvas.Children.Add(p);


         
        }



        /// <summary>
        /// Redraws series any time the Viewport rect changes
        /// </summary>
        protected override void ViewportRectChangedOverride(Rect oldViewportRect, Rect newViewportRect)
        {
            base.ViewportRectChangedOverride(oldViewportRect, newViewportRect);
            this.RenderSeries(false);
        }


        /// <summary>
        /// Redraws series any time the Window rect changes
        /// </summary>
        protected override void WindowRectChangedOverride(Rect oldWindowRect, Rect newWindowRect)
        {
            base.WindowRectChangedOverride(oldWindowRect, newWindowRect);
            this.RenderSeries(false);
        }



        /// <summary>
        /// Overrides property update of this series
        /// </summary>
        protected override void PropertyUpdatedOverride(object sender, string propertyName, object oldValue, object newValue)
        {
            base.PropertyUpdatedOverride(sender, propertyName, oldValue, newValue);
         

            switch (propertyName)
            {
                case ItemsSourcePropertyName:
                case ItemsMemberPathXPropertyName:
                case ItemsMemberPathYPropertyName:
                    this.ItemsTree = new KDTree<object>(2);
                    if (this.ItemsSource == null)
                    {
                        this.RenderSeries(false);
                        return;
                    }
                 
                    foreach (var item in this.ItemsSource)
                    {
                        var dataItem = GetDataPoint(item);
                        this.ItemsTree.Add(item, dataItem.X, dataItem.Y);
                    }
                    
                    this.RenderSeries(false);
                    if (this.XAxis != null)
                    {
                        this.XAxis.UpdateRange();
                    }
                    if (this.YAxis != null)
                    {
                        this.YAxis.UpdateRange();
                    }
                    break;

                case XAxisPropertyName:
                    if (oldValue != null)
                    {
                        ((Axis)oldValue).DeregisterSeries(this);
                    }

                    if (newValue != null)
                    {
                        ((Axis)newValue).RegisterSeries(this);
                    }

                    if ((XAxis != null && !XAxis.UpdateRange()) ||
                        (newValue == null && oldValue != null))
                    {
                        RenderSeries(false);
                    }
                    break;


                case YAxisPropertyName:
                    if (oldValue != null)
                    {
                        ((Axis)oldValue).DeregisterSeries(this);
                    }

                    if (newValue != null)
                    {
                        ((Axis)newValue).RegisterSeries(this);
                    }
                    if ((YAxis != null && !YAxis.UpdateRange()) ||
                        (newValue == null && oldValue != null))
                    {
                        RenderSeries(false);
                    }
                    break;
            }
        }

        /// <summary>
        /// Calculates range of a given axis based on  X/Y values of data items
        /// </summary>
       protected override AxisRange GetRange(Axis axis)
        {
            if (this.ItemsSource == null)
            {
                return base.GetRange(axis);
            }
            if (axis == this.XAxis)
            {
                var min = double.MaxValue;
                var max = double.MinValue;
                foreach (var dataItem in this.ItemsSource)
                {
                    var dataPoint = GetDataPoint(dataItem);
                    min = Math.Min(min, dataPoint.X);
                    max = Math.Max(max, dataPoint.X);
                }
                return new AxisRange(min, max);
            }
            if (axis == this.YAxis)
            {
                var min = double.MaxValue;
                var max = double.MinValue;
                foreach (var dataItem in this.ItemsSource)
                {
                    var DataPoint = GetDataPoint(dataItem);
                    min = Math.Min(min, DataPoint.Y);
                    max = Math.Max(max, DataPoint.Y);
                }
                return new AxisRange(min, max);
            }
            
            return base.GetRange(axis);
        }
        
    
        #endregion

        #region Help Methods
        
        public Point GetSeriesScaledPosition(Point dataPoint)
       {
           var windowRect = this.SeriesViewer.ActualWindowRect;
           var viewport = this.Viewport;

           var xParams = new ScalerParams(windowRect, viewport, this.XAxis.IsInverted);
           xParams.EffectiveViewportRect = this.SeriesViewer.EffectiveViewport;

           var yParams = new ScalerParams(windowRect, viewport, this.YAxis.IsInverted);
           yParams.EffectiveViewportRect = this.SeriesViewer.EffectiveViewport;

           var x = this.XAxis.GetScaledValue(dataPoint.X, xParams);
           var y = this.YAxis.GetScaledValue(dataPoint.Y, yParams);

           return new Point(x, y);
       }
      
        public Point GetSeriesExactPosition(Point world)
       {
           var windowRect = this.SeriesViewer.ActualWindowRect;
           var viewport = this.Viewport;

           var xParams = new ScalerParams(windowRect, viewport, this.XAxis.IsInverted);
           xParams.EffectiveViewportRect = this.SeriesViewer.EffectiveViewport;

           var yParams = new ScalerParams(windowRect, viewport, this.YAxis.IsInverted);
           yParams.EffectiveViewportRect = this.SeriesViewer.EffectiveViewport;

           double x;
           if (this.XAxis.IsVertical)
           {
               x = world.Y;
               x = viewport.Top + viewport.Height * (x - windowRect.Top) / windowRect.Height;
           }
           else
           {
               x = world.X;
               x = viewport.Left + viewport.Width * (x - windowRect.Left) / windowRect.Width;
           }
           double y;
           if (this.YAxis.IsVertical)
           {
               y = world.Y;
               y = viewport.Top + viewport.Height * (y - windowRect.Top) / windowRect.Height;
           }
           else
           {
               y = world.X;
               y = viewport.Left + viewport.Width * (y - windowRect.Left) / windowRect.Width;
           }

           x = this.XAxis.GetUnscaledValue(x, xParams);
           y = this.YAxis.GetUnscaledValue(y, yParams);


           return new Point(x, y);
       }
     
        private DataPoint GetDataPoint(object dataItem)
        {
            var x = (double)dataItem.GetType().GetProperty(this.ItemsMemberPathX).GetValue(dataItem, null);
            
            var y = (double)dataItem.GetType().GetProperty(this.ItemsMemberPathY).GetValue(dataItem, null);
            return new DataPoint(x, y);
        }

        #endregion
        

    }




    internal class DataPoint
    {
        public DataPoint()
        {
            
        }

        public DataPoint(double x, double y)
        {
            X = x;
            Y = y;
        }
        /// <summary>
        /// Gets or sets X
        /// </summary>
        public double X { get; set; }

        /// <summary>
        /// Gets or sets Y
        /// </summary>
        public double Y { get; set; }

        /// <summary>
        /// Gets or sets Point
        /// </summary>
        public Point Point { get { return new Point(X, Y);} }

    }


    

}